program trendtable_mX47;
{
 Cochran-Armitage test for trend.
 Version for input of 2x3 tables.

 History:
 15 Jul 2003 Originally written
 20 Jul 2006 Slightly updated
 20 Apr 2007 Slightly updated
 24 Aug 2014 Change from eq (15.1) to (15.6) in Armitage (2002)
 01 Dec 2021 adapt and test to maXbox4
}

const
 stversion = '24 Aug 2014';

//label 50,99;

var
 numzero,ii,jj,kk:integer;
 table:array[1..2] of array[1..3] of real;
 hold,rr,rn,srx,snx,snxx,chi2:real;


VAR
 rp,x,q,logQ,aln10:real;
 

 procedure initialize;
 begin
  aln10:=LN(10);
  rp:=0.9189385332046727417803296; { LN SQRT(2pi) }
 end; {initialize}

 
 CONST numit=400;
 
 PROCEDURE contfrac; {For x>1}
 { Formula 26.2.14 in Abramowitz and Stegun, 1968, page 932}
 VAR
  i:integer;
  inter:real;
 BEGIN
  inter:=x+numit+1;
  FOR i:=1 TO numit DO inter:=(numit-i+1)/inter+x;
  logQ:=-LN(inter)-0.5*SQR(x)-rp;
  IF x<20 THEN q:=EXP(logQ) ELSE q:=0;
  logQ:=logQ/ln10;
 END; {contfrac}


 procedure pnor(xx:real); {For x<1}
 {Implements formula 26.2.12, page 932, in Abramowitz and Stegun}
 var
  ff,sum,count,x2,two:real;
 begin
  two:=2;
  ff:=xx;
  x2:=SQR(xx);
  count:=1;
  sum:=ff;
  repeat
   count:=count+two;
   ff:=ff*(x2/count);
   sum:=sum+ff;
  until ff<=1.0E-300; {until it underflows}
  q:=0.5-sum*EXP(-0.5*x2-rp);
  logQ:=LN(q)/ln10;
 end;


 PROCEDURE outputstat(xori: real); //unused
 BEGIN
  if (xori<0) then q:=1-q;
  //pnorm:=q;
 END;


 procedure inputX(xori: real);
 begin
  x:=ABS(xori);
 end; {inputX}

function pnorm(xori:real):real;
{
 Copyright (C) Jurg Ott 1988-2002.  Partly programmed by Joe
 Terwilliger using continued fractions.  Based on several formulas in
 Abramowitz and Stegun, "Handbook of mathematical functions, Dover, May 1968.
 Calculates the upper tail probability of the normal
 distribution for a given normal deviate, x.
 If on input x>0 then log(Q) will be output, otherwise log(P) will be output.
  4 Jan 2001  Last program version
 16 Aug 2002  Turned into function to replace old pnorm
}

BEGIN   {pnorm}
 initialize;
 inputX(xori);
 if (ABS(x-0.5)<0.00001)
 then begin
  q:=0.5;
 end
 else if x>1 then contfrac else pnor(x);
 //out;
 if (xori<0) then q:=1-q;
  result:=q;
END; {pnorm}



FUNCTION chipr(x2:REAL; ndf:INTEGER):REAL;
{INCLUDE file for Linkage Utility programs.
 Computes the error probability, chipr = 1-f, where f is the
 chi-square distribution function with ndf degrees of freedom
 evaluated at x2.       Function required: pnorm.
 Follows formulas 26.4.4 and 26.4.21 in Abramowitz and Stegun,
 "Handbook of mathematical functions", Dover, New York 1968}

VAR
 z,p,x,sum,re,ch,chp:REAL;
 i,n1,n2:INTEGER;

BEGIN
 IF (x2>0.0) AND (ndf>0)
 THEN
  IF ndf=1
  THEN BEGIN {ndf=1}
   x:=sqrt(x2);
   result:=2.0*pnorm(x);
  END   (* then *)
  ELSE  {ndf>1}
    IF ndf=2
    THEN result:=exp(-0.5*x2) {ndf=2}
    ELSE BEGIN               {ndf>2}
     n1:=(ndf-1) DIV 2;
     n2:=(ndf-2) DIV 2;
     IF n1=n2
     THEN BEGIN              {ndf is even and >2}
      sum:=1.0;
      re:=0.5*x2;
      ch:=1.0;
      FOR i:=1 TO n2 DO BEGIN
        ch:=ch*re/i;
        sum:=sum+ch;
       END;(* for i *)
      result:=exp(-re)*sum;
     END(* then *)
     ELSE BEGIN              {ndf is odd and >1}
      ch:=sqrt(x2);
      z:=0.39894228*exp(-0.5*x2);
      p:=pnorm(ch);
      IF ndf=3
      THEN result:=2.0*(p+z*ch) {ndf=3}
      ELSE BEGIN               {ndf odd and >3}
       chp:=ch;
       re:=1.0;
       FOR i:=2 TO n1 DO BEGIN
         re:=re+2.0;
         chp:=chp*x2/re;
         ch:=ch+chp;
        END;(* for i *)
       result:=2.0*(p+z*ch);
      END; (*else *)
     END; (* else *)
    END (* else *)
  ELSE result:=1.0;
END;    {chipr}



 procedure Intro;
 begin
  writeln('Program TRENDTABLE version '+(stversion));
  writeln('');
  writeln('For a 2x3 table with 2 rows and 3 columns (SNP genotypes A/A, A/B, B/B),');
  writeln('program carries out Cochran-Armitage test for trend in proportions');
  writeln('across columns. Based on Armitage/Berry/Matthews, "Statistical Methods');
  writeln('in Medical Research" 4th ed., Blackwell, 2002. Example data:');
  writeln('   19  29  24');
  writeln('  497 560 269  furnishes chi-square = 7.1876, p = 0.0073');

 end;
 
//{$I logo.p} 
PROCEDURE logo;
BEGIN
 writeln('');
 writeln('  .-----------------------------------------.');
 writeln('  | STATISTICAL GENETICS UTILITY PROGRAMS   |');
 writeln('  | Copyright (C) 1988-2015        J. Ott   |');
 writeln('  .-----------------------------------------.');
 writeln('');
END; {procedure logo}

procedure setTable;
begin
  table[1][1]:= 19;    table[1][2]:= 29;    table[1][3]:= 24;
  table[2][1]:= 497;   table[2][2]:= 560;   table[2][3]:= 269;
end;

begin
 logo;
 Intro;
 //repeat
  //50:
  writeln('');
  writeln('Enter table with 2 rows and 3 columns (-1 quits)');
  setTable;
  
 (* for ii:=1 to 2 do for jj:=1 to 3 do begin
   {$I-}
   read(table[ii,jj]);
   {$I+}
   if IOresult<>0 then begin
    readln;
    writeln('=== Input error ===');
    goto 50;
   end;
   if table[ii,jj]<0 then goto 99;
  end;
  readln;    *)

  rr:=0.0;
  for kk:=1 to 3 do rr:=rr+table[1][kk];
  rn:=0.0;
  for ii:=1 to 2 do for kk:=1 to 3 do rn:=rn+table[ii][kk];
  srx:=0.0;  snx:=0.0;  snxx:=0.0;
  numzero:=0;
  for kk:=1 to 3 do begin
   srx:=srx+table[1][kk]*kk;
   hold:=table[1][kk]+table[2][kk];
   if hold=0 then inc(numzero);
   snx:=snx+hold*kk;
   snxx:=snxx+hold*kk*kk;
  end; {for kk}
  if numzero>1 then chi2:=-9.9999 else begin
   { Equation (15.1) }
   chi2:=rn*sqr(rn*srx-rr*snx)/( rr*(rn-rr)*(rn*snxx-sqr(snx)) );
   { Equation (15.6) }
   chi2:=(rn-1.0)*chi2/rn;
  end;
  write(' Chi2(1) = '+flots(chi2)); //:6:4);
  if numzero<2 then writeln(' , p ='+flots(chipr(chi2,1))) {:9:6)} else writeln('Nothing');
  if numzero=1 then writeln(' Warning: only 2 genotypes present');
  if numzero=2 then writeln(' only 1 genotype present');

 //until false;
 //99:

End.

Enter table with 2 rows and 3 columns (-1 quits)
 Chi2(1) = 7.187604555372988
 , p =0.007340891610159
 mX4 executed: 29/08/2022 17:14:44  Runtime: 0:0:1.587  Memload: 42% use
 
    
